import ctypes
import os
import threading

from textual.app import App, ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.widgets import Header, Footer, TabbedContent, TabPane, Button, Input, RichLog, Static
from textual.screen import ModalScreen
from textual import on, events, work
from rich.style import Style
from rich.text import Text

# Global variables
FIFO_TRANSACTION = "/tmp/transFIFO"
FIFO_CLIENT = "/tmp/fifo" + str(os.getpid())
server_fifo = open(FIFO_TRANSACTION, 'wb')
read_from_server = True
exit_event = threading.Event()

# Create the C struct to send with the pipe with the fields int pid and char transaction[100]
class Transaction(ctypes.Structure):
    _fields_ = [("pid", ctypes.c_int),
                ("transaction", ctypes.c_char * 100)]    

def to_utf8(string) -> bytes:
    """Convert a string to utf-8 encoding for the fifo C module"""
    return string.encode('utf-8')
    
def get_transaction(type_and_args: str) -> Transaction:
    """Return a Transaction object with the pid and transaction fields filled"""
    return Transaction(os.getpid(), to_utf8(type_and_args))


def send_transaction(transaction: Transaction):
    """Send a transaction to the server and wait for a response"""
    global server_fifo
    server_fifo.write(transaction)
    server_fifo.flush()
    
class MyApp(App):
    BINDINGS = [
        ("ctrl+t", "toggle_dark", "Toggle dark mode"),
        ("ctrl+q", "quit", "Quit")
        ]
    TITLE = "Gestion VMS"
    CSS_PATH = "main.tcss"

    # Create the client fifo
    try:
        os.remove(FIFO_CLIENT)
    except OSError:
        pass
    os.mkfifo(FIFO_CLIENT)

    def compose(self) -> ComposeResult:
        yield Header()
        yield Horizontal(
            Client(classes=("container")),
            RichLog(classes=("container logger"))
        )
        yield Footer()

    def action_toggle_dark(self) -> None:
        """An action to toggle dark mode."""
        self.dark = not self.dark

    def on_mount(self) -> None:
        """Called when the application is mounted."""
        # Call a worker function to read the server response
        self.logger = self.query_one(RichLog)
        self.logger.border_title = "Reception"

    def on_ready(self) -> None:
        """Called  when the DOM is ready."""
        # Call a worker function to read the server response
        self.read_server_response()

    @work(name="receiver", thread=True)
    def read_server_response(self) -> None:
        """Read the server response and display it in the log"""
        global FIFO_CLIENT
        # Open the fifo non-blockingly and read the response
        with os.fdopen(os.open(FIFO_CLIENT, os.O_RDONLY | os.O_NONBLOCK), 'rb') as fifo:
            while not exit_event.is_set():
                response = fifo.read()
                if response:
                    # Remove the nul bytes from the response
                    response = response.replace(b'\x00', b'')
                    self.write_response(response.decode('utf-8'))

        # Remove the fifo when the thread is done
        print("Removing fifo")
        os.remove(FIFO_CLIENT)
        print("Exiting receiver thread")

    def write_response(self, response: str) -> None:
        """Write the response to the log"""
        # Get the type of the response (ex. A, E, L, X)
        type = response.split(" ")[0]        

        # Style the response depending on the type
        if type == "A":
            style = Style(color="green")
            # Remove the type from the response
            response = response[2:]
        elif type == "E":
            style = Style(color="red")
            response = response[2:]
        elif type == "L":
            style = Style(color="blue")
            response = response[2:]
        elif type == "X":
            style = Style(color="yellow")
            response = response[2:]
        else:
            style = Style(color="white")
        
        # Write the response to the log
        self.logger.write(Text(response, style=style))
        

    @on(events.Shutdown)
    def on_shutdown(self, event: events.Shutdown) -> None:
        """Close the fifo and exit the program"""
        self.quit()


    def action_quit(self) -> None:
        """Close the fifo and exit the program"""
        exit_event.set()
        global server_fifo
        server_fifo.close()
        self.app.exit()
    

    def warn(self, message: str) -> None:
        """Write a warning to the log."""
        self.push_screen(Warning(message))

class Warning(ModalScreen):
    DEFAULT_CSS = """
    .warning {
        padding: 2;
        color: orange;
        border: tall solid orange;
        background: black;
        align: center middle;
        text-align: center;
        width: 50%;
    }
    
    Vertical{
        align: center middle;
        background: black;
        height: 100%;
        width: 100%;
    } 
    
    #ok {
        width: 50%;
        background: black;
        border: tall solid orange;
        align: center middle; 
    }
    
    #ok:hover {
        color: orange;
    }

    """

    def __init__(self, message: str) -> None:
        super().__init__()
        self.message = message


    def compose(self) -> ComposeResult:
        yield Container(
            Vertical(
                Static(self.message, classes="warning"),
                Button("Ok", name="ok", id="ok"),
                classes="scrollable"
            )
        )
    @on(Button.Pressed, "#ok")
    def ok(self) -> None:
        """Close the modal."""
        self.dismiss()

class Client(Container):
    BINDINGS = [
        ("a", "show_tab('add')", "Ajouter"),
        ("e", "show_tab('delete')", "Eliminer"),
        ("l", "show_tab('list')", "Lister"),
        ("x", "show_tab('execute')", "Executer")
    ]
    BORDER_TITLE = "Client"

    def compose(self) -> ComposeResult:
        with TabbedContent():
            with TabPane(title="Ajouter", id="add"):
                yield AddVM()
            with TabPane(title="Eliminer", id="delete"):
                yield DeleteVM()
            with TabPane(title="Lister", id="list"):
                yield ListVM()
            with TabPane(title="Executer", id="execute"):
                yield Execute()

class AddVM(Vertical):
    DEFAULT_CSS = """
   #add:hover 
   {
    background: green;
    outline: green;
   }
    """
    def compose(self) -> ComposeResult:
        yield Container(
            Static("Ajouter une VM", classes="space_below"),
            Button("Ajouter", name="add", id="add"),
            classes="scrollable"
        )
    @on(Button.Pressed, "#add")
    def add_vm(self) -> None:
        """Send a transaction to the server to add a VM"""
        send_transaction(get_transaction("A"))


class DeleteVM(Vertical):
    DEFAULT_CSS = """
   #delete:hover 
   {
    background: red;
    outline: red;
   }
    """
    def compose(self) -> ComposeResult:
        yield Container(
            Static("Entrer le chiffre correspondant à la VM à éliminer", classes="space_below"),
            Input(placeholder="Ex. 1", id="Idelete"),
            Button("Eliminer la VM", name="delete", id="delete"),
            classes="scrollable"
        )

    def __init__(self) -> None:
        super().__init__()
        self.warn = self.app.warn
    
    def on_mount(self) -> None:
        """Set the input when the component is mounted."""
        self.input = self.query_one(Input)

    @on(Button.Pressed, "#delete")
    @on(Input.Submitted, "#Idelete")
    def delete_vm(self) -> None:
        """Send a transaction to the server to delete a VM"""
        # check if the input is a number
        if self.input.value.isnumeric():
            send_transaction(get_transaction("E " + self.input.value))
        else:
            self.warn("Entrer un chiffre valide (ex. 1)")


class ListVM(Vertical):
    DEFAULT_CSS = """
   #list:hover 
   {
    background: blue;
    outline: blue;
   }
    """
    def compose(self) -> ComposeResult:
        yield Container(
            Static("Entrer le range correspondant aux VM à lister", classes="space_below"),
            Input(placeholder="Ex. 1-4", id="Ilist"),
            Button("Lister les VM", name="list", id="list"),
            classes="scrollable"
        )

    def __init__(self) -> None:
        super().__init__()
        self.warn = self.app.warn
    
    def on_mount(self) -> None:
        """Set the input when the component is mounted."""
        self.input = self.query_one(Input)

    @on(Button.Pressed, "#list")
    @on(Input.Submitted, "#Ilist")
    def list_vm(self) -> None:
        """Send a transaction to the server to list VMs"""
        # check if the input is a range (ex. 1-4)
        if "-" in self.input.value and len(self.input.value.split("-")) == 2:
            range = self.input.value.split("-")
            if range[0].isnumeric() and range[1].isnumeric():
                send_transaction(get_transaction("L " + self.input.value))
            else:
                self.warn("Entrer un range valide (ex. 1-4)")
        else:
            self.warn("Entrer un range valide (ex. 1-4)")


class Execute(Vertical):
    DEFAULT_CSS = """
   #execute:hover 
   {
    background: yellow;
    outline: yellow;
   }
    """
    def compose(self) -> ComposeResult:
        yield Container(
            Static("Entrer le numéro de la VM et le nom du fichier à exécuter", classes="space_below"),
            Input(placeholder="Ex. 1 Mult-Numbers.olc3", id="Iexecute"),
            Button("Send a transaction", name="execute", id="execute"),
            classes="scrollable"
        )

    def __init__(self) -> None:
        super().__init__()
        self.warn = self.app.warn
        
    def on_mount(self) -> None:
        """Set the input when the component is mounted."""
        self.input = self.query_one(Input)

    @on(Button.Pressed, "#execute")
    @on(Input.Submitted, "#Iexecute")
    def execute(self) -> None:
        """Send a transaction to the server to execute a command"""
        # Check if the input a vm number and a file name
        if self.input.value.split(" ")[0].isnumeric() and ".olc3" in self.input.value.split(" ")[1]:
            send_transaction(get_transaction("X " + self.input.value))
        else:
            self.warn("Entrer un fichier valide (ex. Mult-Numbers.olc3)")



if __name__ == '__main__':
    app = MyApp()
    app.run()
    print("Exiting main application")

